import os
import numpy as np
import DTData 
from support_modules import DTError, DTOutput

class DTAnalysis:
        
    
    def __init__(self, output_location = 'delineate_output', output_filename_stem=None, output_file_types=None, nits=None, xval_type=None, nfolds=None,
     train_val_test=None, scaling_method=None, scaling=None, classify_over=None, backend_options=None, dataset=None, model=None, job_struct=None, job_file_hash = None, xval_sa = None, pre_existing_file_behavior = 'silent_append' ):
        self.output_location = output_location
        self.output_filename_stem = output_filename_stem
        self.output_file_types = output_file_types
        self.nits = nits
        self.xval_type = xval_type
        self.nfolds = nfolds
        self.train_val_test = train_val_test
        self.scaling_method = scaling_method
        self.scaling = scaling
        self.classify_over = classify_over
        self.backend_options = backend_options
        self.dataset = dataset
        self.model = model
        self.sync_backend_options_with_model(warn_if_none=False)
        self.job_struct = job_struct
        self.job_file_hash = job_file_hash
        self.xval_sa = xval_sa
        self.pre_existing_file_behavior = pre_existing_file_behavior
        # TK: insert appropriate error checking here
    
    
    def run(self):
        # TK: insert appropriate error checking here
        #  specifically, make sure all necessary attributes exist and are the right data type, etc.
        
        self.sync_backend_options_with_model(warn_if_none=False)
        
        if isinstance( self.nits, list ):
            nits = self.nits[0]
        else:
            nits = self.nits
        
        for iteration_num in range(nits):
            print('DeLINEATE: Iteration {0}'.format(iteration_num+1))
            if self.xval_type == 'transfer_over_sa':
                self.run_transfer( iteration_num )
            elif self.xval_type == 'single':
                self.run_single( iteration_num )
            elif self.xval_type == 'loop_over_sa':
                self.run_loop_over_sa( iteration_num )
            else:
                DTError.err("DTAnalysis: Unknown 'xval_type' parameter specified!")
    
    
    def run_transfer(self, iteration_num):
        from keras.models import load_model
        unique_sa_values = np.unique(self.dataset.sa[self.xval_sa])
        for sa_value in unique_sa_values:
            self.dataset.mask_sa(self.xval_sa, np.setdiff1d(unique_sa_values, sa_value))
            [self.train_data, self.val_data, self.test_data, self.train_labels, self.val_labels, self.test_labels] = self.dataset.slice_train_val_test( .7, .3, 0, self.classify_over )

            if self.scaling != None:
                self.rescale_data()

            [self.train_labels, self.val_labels, self.test_labels] = DTData.DTData.train_val_test_to_categorical( self.train_labels, self.val_labels, self.test_labels )
            self.model.build()
            self.model.train( self.train_data, self.val_data, self.train_labels, self.val_labels )
            self.model.model.save("./transfer_temp")

            self.dataset.mask_sa(self.xval_sa, sa_value)

            print("Trained for subject " + str(sa_value))
            #TK: Explicit check to see if nits is list
            for i in range(self.nits[1]):

                #Probably doesn't work with tensorflow. 
                #TK: Update to keras.models.clone_model(model, input_tensors) when we updated Keras to at least 2.0.7
                self.model.model = load_model("./transfer_temp")
                [self.train_data, self.val_data, self.test_data, self.train_labels, self.val_labels, self.test_labels] = self.dataset.slice_train_val_test( self.train_val_test[0], self.train_val_test[1], self.train_val_test[2], self.classify_over )
                if self.scaling != None:
                    self.rescale_data()

                [self.train_labels, self.val_labels, self.test_labels] = DTData.DTData.train_val_test_to_categorical( self.train_labels, self.val_labels, self.test_labels )
                self.model.train( self.train_data, self.val_data, self.train_labels, self.val_labels )
                [self.test_acc, self.test_loss, self.test_prediction_scores] = self.model.test( self.test_data, self.test_labels )
                print("RESULT FOR " + str(sa_value) + " ITERATION " + str(i) + " IS " + str(self.test_acc))
                
                fold_to_write = '{0},{1}'.format(sa_value, i)
                if iteration_num == 0 and sa_value == unique_sa_values[0] and i == 0:
                    self.write_output( iteration_num, fold=fold_to_write, do_headers=True, do_config=True)
                else:
                    self.write_output( iteration_num, fold=fold_to_write )
    
    
    def run_loop_over_sa(self, iteration_num):
        unique_sa_values = np.unique(self.dataset.sa[self.xval_sa])
        for sa_value in unique_sa_values:
            print('DeLINEATE: Looping over SA values; current fold value is {0}'.format(sa_value))
            self.dataset.mask_sa(self.xval_sa, sa_value)
            
            [self.train_data, self.val_data, self.test_data, self.train_labels, self.val_labels, self.test_labels, self.test_indices] = self.dataset.slice_train_val_test( self.train_val_test[0], self.train_val_test[1], self.train_val_test[2], self.classify_over, return_test_indices=True )
            
            if self.scaling != None:
                self.rescale_data()
            if self.model.backend == 'keras':
                [self.train_labels, self.val_labels, self.test_labels] = DTData.DTData.train_val_test_to_categorical( self.train_labels, self.val_labels, self.test_labels )
            
            self.model.build()
            self.model.train( self.train_data, self.val_data, self.train_labels, self.val_labels )
            [self.test_acc, self.test_loss, self.test_prediction_scores] = self.model.test( self.test_data, self.test_labels )
            
            if iteration_num == 0 and sa_value == unique_sa_values[0]:
                self.write_output( iteration_num, fold=sa_value, do_headers=True, do_config=True)
            else:
                self.write_output( iteration_num, fold=sa_value )
            
            print('\n')
    
    
    def run_single(self, iteration_num):
        [self.train_data, self.val_data, self.test_data, self.train_labels, self.val_labels, self.test_labels, self.test_indices] = self.dataset.slice_train_val_test( self.train_val_test[0], self.train_val_test[1], self.train_val_test[2], self.classify_over, return_test_indices=True )
        
        if self.scaling != None:
            self.rescale_data()
        if self.model.backend == 'keras':
            [self.train_labels, self.val_labels, self.test_labels] = DTData.DTData.train_val_test_to_categorical( self.train_labels, self.val_labels, self.test_labels )
        
        self.model.build()
        self.model.train( self.train_data, self.val_data, self.train_labels, self.val_labels )
        [self.test_acc, self.test_loss, self.test_prediction_scores] = self.model.test( self.test_data, self.test_labels )
        
        if iteration_num == 0:
            self.write_output( iteration_num, do_headers=True, do_config=True)
        else:
            self.write_output( iteration_num )
            
        print('\n')
    
    
    def sync_backend_options_with_model(self, warn_if_none=True, warn_if_different=True):
        if self.model is None:
            if warn_if_none:
                print("Warning: DTAnalysis.sync_backend_options_with_model() was called, but self.model is None! Not doing anything here.")
                print(" ")
            return
        
        if self.model.backend_options is None:
            self.model.backend_options = self.backend_options
            return
        
        if self.model.backend_options != self.backend_options:
            self.model.backend_options = self.backend_options
            if warn_if_different:
                print("Warning: DTAnalysis.sync_backend_options_with_model() was called, and the model's backend options were different!")
                print("         This is OK if you meant to update the model with new backend options from your DTAnalysis object, but ")
                print("         you might need to look into it further if this message is unexpected to you.")
    
    
    #TK: maybe change into a DTData classmethod like PyMVPA wrapping
    def rescale_data(self):
        if (self.scaling is None) and (self.scaling_method is None):
            return #should have already been checked but you never know
        
        #for the sake of backwards compatibility, if scaling is specified with no method assume percentile
        if (self.scaling_method == "percentile") or (self.scaling_method is None):
            if self.scaling >= 100 or self.scaling <= 1:
                DTError.err("DTAnalysis: Scaling needs to be >1 and <100 if you are going to use it (representing a percentile)")
            
            scale_value = np.percentile(np.abs(self.train_data), self.scaling)
            self.train_data = self.train_data / scale_value
            self.val_data = self.val_data / scale_value
            self.test_data = self.test_data / scale_value
            
        elif self.scaling_method == "standardize":
            training_mean = np.mean(self.train_data)
            training_std = np.std(self.train_data)
            self.train_data = (self.train_data - training_mean)/training_std
            self.val_data = (self.val_data - training_mean)/training_std
            self.test_data = (self.test_data - training_mean)/training_std
            
        elif self.scaling_method == "map_range":
            if self.scaling is None:
                targ_min = 0
                targ_max = 1
            else:
                targ_min = self.scaling[0]
                targ_max = self.scaling[1]
            training_min = np.min(self.train_data)
            training_max = np.max(self.train_data)
            max_min_dist = training_max - training_min
            
            norm_train_data = (self.train_data - training_min)/max_min_dist
            self.train_data = norm_train_data * (targ_max-targ_min) + targ_min
            norm_val_data = (self.val_data - training_min)/max_min_dist
            self.val_data = norm_val_data * (targ_max-targ_min) + targ_min
            norm_test_data = (self.test_data - training_min)/max_min_dist
            self.test_data = norm_test_data * (targ_max-targ_min) + targ_min
            
        elif self.scaling_method == "mean_center":
            training_mean = np.mean(self.train_data)
            self.train_data = self.train_data - training_mean
            self.val_data = self.val_data - training_mean
            self.test_data = self.test_data - training_mean
            
        else:
            DTError.err("DTAnalysis: Scaling method '{0}' is not currently supported. Implemented options are: percentile, standardize, map_range, and mean_center.".format(str(self.scaling_type)))
    
    
    def write_output(self, iteration, fold=0, do_headers=False, do_config=False):
        if self.output_file_types is None:
            return
        
        if self.output_filename_stem is None:
            self.output_filename_stem = os.path.join(self.output_location,self.job_file_hash)
            print("No output filename stem was specified, using md5 hash of the job file.")
        
        if not isinstance( self.output_file_types, list):
            self.output_file_types = [self.output_file_types]
            
        if not os.path.isdir(self.output_location):
            os.makedirs(self.output_location)
        
        for this_output_file_type in self.output_file_types:
            if this_output_file_type == "acc_summary":
                self.write_output_acc_summary( iteration, fold, do_headers )
                
            elif this_output_file_type == "scores":
                self.write_output_scores( iteration, fold, do_headers )
                
            elif this_output_file_type == "labels":
                self.write_output_labels( iteration, fold, do_headers )
                
            elif this_output_file_type == "madame_kerasana":
                self.write_output_kerasana( iteration, fold, do_headers )
            
            elif this_output_file_type[0:5] == 'tags:':
                this_tag = this_output_file_type[5:]
                self.write_output_tags( iteration, this_tag, fold, do_headers )
                
            elif this_output_file_type == "job_config":
                if do_config:
                    output_filename = os.path.join(self.output_location, self.output_filename_stem + "_job_config.json")
                    DTOutput.write_job_config(output_filename, self.job_struct, self.pre_existing_file_behavior)
            else:
                DTError.err("DTAnalysis.write_output(): Unsupported output type")
    
    
    def write_output_acc_summary(self, iteration, fold=0, do_headers=False):
        output_filename = os.path.join(self.output_location, self.output_filename_stem  + "_accs.tsv")
        
        if do_headers:
            DTOutput.write_headers( output_filename, ['Iteration','Fold','MeanAcc','MeanLoss'], self.pre_existing_file_behavior )
        
        DTOutput.write_one_row_values( output_filename, ['{0:05d}', '{0}',    '{0:.9f}',      '{0:.9f}'], 
                                                        [iteration,  fold,  self.test_acc, self.test_loss] )
        
    def write_output_kerasana(self, iteration, fold=0, do_headers=False):
        import random
        random.seed(int(self.test_acc*100))
        output_filename = os.path.join(self.output_location, self.output_filename_stem  + "_madame_kerasana.tsv")
        the_future                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              = ["It is certain","It is decidedly so","Without a doubt","Yes definitely","You may rely on it","As I see it, yes","Most likely","Outlook good","Yes","Signs point to yes","Reply hazy try again","Ask again later","Better not tell you now","Cannot predict now","Concentrate and ask again","Don't count on it","My reply is no","My sources say no","Outlook not so good","Very doubtful"]
        a_glimpse                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           = the_future[random.randrange(len(the_future))]
        
        if do_headers:
            DTOutput.write_headers( output_filename, ['Iteration','Fold','Prediction'], self.pre_existing_file_behavior )
        
        DTOutput.write_one_row_values( output_filename, ['{0:05d}', '{0}',    '{0}'], 
                                                        [iteration,  fold,  a_glimpse] )
    
    
    def write_output_scores(self, iteration, fold=0, do_headers=False):
        self.n_scores = len(self.test_prediction_scores)
        self.n_classes = len(np.unique(self.test_labels))
        if (self.model.backend=='pymvpa'):
            if (self.n_classes > 2) and ('SVM' in self.model.model.summary()):
                self.write_output_scores_pymvpa_multiclass(iteration, fold, do_headers)
            else:
                self.write_output_scores_generic(iteration, fold, do_headers)
        else:
            self.write_output_scores_generic(iteration, fold, do_headers)
    
    
    def write_output_scores_generic(self, iteration, fold, do_headers):
        output_filename = os.path.join(self.output_location, self.output_filename_stem + "_scores.tsv")
        
        if do_headers:
            DTOutput.write_headers( output_filename, ['Iteration','Fold','Class', {'base':'Score','length':self.n_scores}  ], self.pre_existing_file_behavior )
        
        scores_to_write = list(map(list,zip(*self.test_prediction_scores)))
        # this is a super-gross line of code, but it should work for either Python 2 or 3, and also for either numpy arrays or a list of Python lists
        # see: https://stackoverflow.com/questions/6473679/transpose-list-of-lists for where it came from
        # it looks like self.test_prediction_scores can either end up being a numpy array or a list of lists? So we are stuck doing this unless we want 
        #   to enforce one format over the other when that variable gets assigned
        # anyway, we need to transpose the scores so we have them as class x score instead of score x class, and this is the only way I know how right now -- MRJ
        
        for class_num in range(len(scores_to_write)):
            DTOutput.write_one_row_values( output_filename, ['{0:05d}', '{0}', '{0:03d}',                  '{0:.9f}'], 
                                                            [iteration,  fold,   class_num, scores_to_write[class_num]] )
    
    
    def write_output_scores_pymvpa_multiclass(self, iteration, fold, do_headers):
        output_filename = os.path.join(self.output_location, self.output_filename_stem + "_scores.tsv")
        
        if do_headers:
            DTOutput.write_headers( output_filename, ['Iteration','Fold','Class 1','Class 2', {'base':'Score','length':self.n_scores}  ], self.pre_existing_file_behavior )
        
        # for reasons I don't want to think about right now we're getting a list of lists of dictionaries. use set operations to ignore the redundant keys
        # as a token kindness to matt I won't even use list comprehension to do it
        comparison_keys = []
        redundant_keys = set()
        for comparison in self.test_prediction_scores[0][0].keys():
            s = frozenset(comparison)
            if s not in redundant_keys:
                redundant_keys.add(s)
                comparison_keys.append(comparison)
        
        for comparison_num in comparison_keys:
            scores_to_write = [val[0][comparison_num] for val in self.test_prediction_scores]
            DTOutput.write_one_row_values( output_filename, ['{0:05d}', '{0}',  '{0:03d}',          '{0:03d}',                 '{0:.9f}'], 
                                                            [iteration,  fold,   comparison_num[0], comparison_num[1],         scores_to_write] )
    
    
    def write_output_labels(self, iteration, fold=0, do_headers=False):
        output_filename = os.path.join(self.output_location, self.output_filename_stem + "_labels.tsv")
        n_labels = len(self.test_labels)
        
        if isinstance(self.test_labels[0], np.ndarray):  # one-hot coding, presumably from Keras
            if do_headers:
                DTOutput.write_headers( output_filename, ['Iteration','Fold','Class', {'base':'Label','length':n_labels}  ], self.pre_existing_file_behavior )
            
            labels_to_write = list(map(list,zip(*self.test_labels)))
            # this is a super-gross line of code, but it should work for either Python 2 or 3, and also for either numpy arrays or a list of Python lists
            # similar situation as above with the output scores... see comment up there        
            
            for class_num in range(len(labels_to_write)):
                DTOutput.write_one_row_values( output_filename, ['{0:05d}', '{0}', '{0:03d}',                  '{0:.9f}'], 
                                                                [iteration,  fold,   class_num, labels_to_write[class_num]] )
        else: # not one-hot coding, presumably from PyMVPA
            if do_headers:
                DTOutput.write_headers( output_filename, ['Iteration','Fold', {'base':'Label','length':n_labels}  ], self.pre_existing_file_behavior ) #no need for 'Class' column
            
            DTOutput.write_one_row_values( output_filename, ['{0:05d}', '{0}',        '{0:.9f}'], 
                                                            [iteration,  fold,   self.test_labels] )
    
    
    def write_output_tags(self, iteration, tag_name, fold=0, do_headers=False):
        output_filename = os.path.join(self.output_location, self.output_filename_stem + "_tags_" + tag_name + ".tsv")
        tags_to_write = self.dataset.sa[tag_name]
        tags_to_write = tags_to_write[self.test_indices]
        
        n_tags = len(tags_to_write)
        
        if do_headers:
            DTOutput.write_headers( output_filename, ['Iteration','Fold', {'base':'Tag','length':n_tags}  ], self.pre_existing_file_behavior )
        
        DTOutput.write_one_row_values( output_filename, ['{0:05d}', '{0}',         '{0}'], 
                                                        [iteration,  fold, tags_to_write] )
        
    
    @classmethod
    def validate_arguments( cls, args ):
        # this is the laziest and dumbest possible version of a method like this, which really should inspect the arguments and give some intelligent
        # warning messages and so forth. But it'll do for now to get us off the ground
        try:
            test_obj = cls( **args )
        except:
            global dt_debug_mode
            if 'dt_debug_mode' not in globals():
                dt_debug_mode = True # TK: change to false when we go into public production
            
            if dt_debug_mode:
                DTError.err( 'DTAnalysis: validate_arguments() failed' )
            else:
                return False
        return True