# The DeLINEATE Toolbox -- basic info #

This file is just a quick overview -- for details on usage of various parts of the toolbox, see the Markdown files within the `documentation` folder of this repository or online at http://delineate.it

In theory these docs should be kept in sync with the documentation on the website, but you never know. If you discover discrepancies, please let the dev team know. Best way is probably to create an Issue using the Issues page of our Bitbucket repository -- this will require you to sign up for a (free) Bitbucket account.


## INSTALLATION / DEPENDENCIES ##

The installation of our toolbox is pretty straightforward, but fair warning: Installation of the toolboxes that we rely upon could be a minor-to-major pain. (We rely on PyMVPA for conventional MVPA, and Keras for deep learning -- and then Keras, in turn, depends on either a TensorFlow or Theano backend. And if you're doing GPU-accelerated learning, which hopefully you are if you'd like your analyses to complete while we're all still alive, you will also need Nvidia's CUDA and cuDNN tools.)

We hope to have a more comprehensive guide soon, but since we're working with open-source tools that are in active development, it's hard to provide a definitive set of steps that will work for everyone and won't go out of date nearly instantaneously. But basically, if you have a setup that is capable of running PyMVPA (http://www.pymvpa.org) and/or Keras (https://keras.io), then you should be able to run our stuff without any additional setup. Those toolboxes' documentation is not amazingly detailed but with a little tech-savviness, plus Google and perseverance, you should be able to get it all working. If you have real troubles, the dev team is happy to help people with their own setups -- all we ask in exchange is your help in fleshing out the documentation, so we can pay it forward to future users.

If it helps at all, we have some super-janky and probably somewhat out-of-date (possibly even straight-up incorrect) notes from our initial attempts to set up our own deep learning workstations. Feel free to crib from there as helpful, but take it all with a huge grain of salt. We hope to migrate the bits from that page that are actually useful and true to this documentation eventually, but for now... caveat emptor: http://matthewrjohnson.net/lab/rapwiki/DeepLearningSetup

See the next section for more specific deets on the kinds of hardware/software you're gonna need to get this stuff all working.


## COMPATIBILITY ##

### Hardware and OS ###

*Overview.* Minimally, just about any computer with a working Python installation will do if you want to try out this toolbox. PyMVPA does not do GPU acceleration at the moment, and Keras will run its analyses on the CPU(s) if it doesn't find a compatible GPU. However, most of the value of this toolbox probably comes from doing GPU-accelerated deep learning analyses. You'll probably want reasonable computing hardware to get decent speed out of whatever you're doing, but most of the other stuff doesn't matter nearly as much as the GPU. So if you're trying to allocate resources on a budget, sink more money into the GPU and less into maxing out the other stuff. A decent amount of system memory is a good idea (16 GB would be OK, 32 GB even better... basically you want RAM equivalent in size to the largest chunk of data you'd like to analyze at once, plus a few GB of headroom for the OS and applications). SSD vs conventional hard drive doesn't matter that much beyond the usual speedup to your system overall... most of the time you'll be loading the data once and then analyzing it for a long time thereafter, so disk reading/writing time will not be most people's speed bottleneck. CPU cores mainly only matter if you want to run Keras (deep learning) analyses without GPU acceleration (which even with many CPUs is going to be much slower than using a GPU, so not super-recommended), or if you want to run a bunch of PyMVPA processes simultaneously. So for most people, a run-of-the-mill CPU with 4 medium-fast cores is fine and will not be your main speed bottleneck either. Again, it all depends on your use case but as long as you have a medium-powerful computer in conventional terms, the main thing to think about will be the GPU.

*GPU.* At time of writing, the deep learning libraries that DeLINEATE depends upon only do GPU acceleration for Nvidia GPUs. So that's easy. For most people, you'll see a huge gain of using any semi-decent GPU versus relying on CPUs, and then relatively modest benefits with increasingly fancy GPUs. A pretty decent gaming GPU (e.g. a GeForce GTX 1060, currently retailing for ~300-400 USD) should be sufficient for many cases. You can go on up the line if you like (the GTX 1070 and 1080 line are also nice GPUs), but for the increase in price, you could be better off building more systems with cheaper GPUs if you want to get more analyses run faster, rather than maxing out each GPU. Again, your mileage may vary. If you're trying to decide, the dev team is happy to consult in exchange for your help fleshing out the documentation and benchmark data of this toolbox.

One key note is that even if there are not huge speed bumps between GPU models, they are also differentiated by the amount of onboard GPU RAM. This makes a difference because some Keras models occupy a LOT of GPU RAM. So if you expect to run a complex neural network model and be able to push large chunks of data through it, you may need to shell out for a nicer GPU purely for RAM purposes, even if the speed boost is not very important to you.

### A personal note from the devs ###

OK, full disclosure. We are pushing to have a public release out before a grant deadline and ALL of the verbiage of the documentation is probably not going to happen before that deadline. We are still actively developing the code and docs either way, but this is kinda the stopping point for writing the big, long, primary README file for now. Stuff below is mostly notes-to-ourselves.

Ditto the stuff in the `documentation` directory and on the website... a lot of it is still in primordial format but if you happen upon this toolbox before those primordial forms get overwritten with more content, please just bear with us. You are welcome to contact the dev team with questions and that will help motivate us regarding which parts should get fleshed out first.


### Code libraries and toolboxen we depend on ###

### Python versions (2? 3?) ###
* Keras should in theory work with both Python 2 and 3, as of the time of writing? (2018-03-20)
* Officially, seems like PyMVPA is Python 2.7 only, but so far we have had good luck with Python 3 in most use cases and have seemingly managed to avoid most of the stuff that would break under Python 3.
* Need to make some notes on Keras / PyMVPA compatible versions, TensorFlow vs Theano, etc.



TERMINOLOGY (major object types, etc.)

