import json, os
from . import DTError, DTUtils

def write_headers( filename, header_spec_list, pre_existing_file_behavior, delimiter="\t"  ):
    # each item in 'header_spec_list' can either be a string or a dict
    # if a string, the header just gets written out literally (with a delimiter)
    # if a dict, it should have two keys: 'base' and 'length'
    
    header_spec_list = DTUtils.listify( header_spec_list ) #turn single item into list if necessary
    fid = check_existing(filename, pre_existing_file_behavior)
        
    this_delim = "" #no delimiter for first item
    for header_spec in header_spec_list:
        if isinstance( header_spec, dict ):
            base = header_spec['base']
            n_items = header_spec['length']
            for i in range( n_items ):
                header_string = base + '{0:07d}'.format( i ) # TK? maybe in future make number of digits optionally configurable; for now it's a pain
                fid.write( this_delim + header_string )
                this_delim = delimiter # a little inefficient if delimiter is big, but it's not likely to be
        else: #assume it's a string
            fid.write( this_delim + header_spec )
            this_delim = delimiter # a little inefficient if delimiter is big, but it's not likely to be
    
    fid.write("\n")
    fid.close()


def write_one_row_values( filename, format_specs, values, delimiter="\t" ):
    # 'format_specs' should be a list of format specification strings, e.g. ["{0:05d}", "{0:.9f}", "{0}"] or whatnot, one string for each item in 'values'
    # 'values' should be a list; each item in the list can either be a single value (string, number, whatever, as long as it matches corresponding format spec),
    #          or a list of such values, in which case each item in the list will be printed as a separate cell
    
    format_specs = DTUtils.listify( format_specs )
    values = DTUtils.listify( values )
    
    if len(format_specs) != len(values):
        DTError.err("DTOutput: Length of 'format_specs' does not match length of 'values' in input arguments!")
    
    fid = open( filename, 'at' ) # for now assume we always append
    this_delim = "" #no delimiter for first item
    for i in range(len(format_specs)):
        this_spec = format_specs[i]
        these_vals = values[i]
        these_vals = DTUtils.listify( these_vals )
        
        for this_val in these_vals:
            fid.write( this_delim + this_spec.format( this_val ) )
            this_delim = delimiter # a little inefficient if delimiter is big, but it's not likely to be
    
    fid.write("\n")
    fid.close()

def write_job_config(filename, job_struct, pre_existing_file_behavior):
    fid = check_existing(filename, pre_existing_file_behavior)
    out_str = json.dumps(job_struct,sort_keys=True,indent=3)
    fid.write(out_str)
    fid.close()
    
def check_existing(filename, pre_existing_file_behavior):
    file_exists = ( os.path.isfile(filename) )
    
    if file_exists and (pre_existing_file_behavior == 'overwrite'):
        fid = open( filename, 'wt' ) 
    else:
        fid = open( filename, 'at' ) 
    
    if file_exists and (pre_existing_file_behavior != 'overwrite'):
        fid.write('\n\n\n\n')
        
    return fid